#include <math.h>
#include "Gdc2.h"
#include "GdcTools.h"

bool IsGreyScale(GSurface *pDC)
{
	bool Status = false;
	if (pDC)
	{
		GPalette *Pal = pDC->Palette();
		Status = true;
		if (Pal AND Pal->GetSize() > 0)
		{
			for (int i=0; i<Pal->GetSize(); i++)
			{
				GdcRGB *p = (*Pal)[i];
				if (p)
				{
					int Grey = i * 255 / Pal->GetSize();
					if (p->R != Grey OR
						p->G != Grey OR
						p->B != Grey)
					{
						Status = false;
					}
				}
			}
		}
	}
	return Status;
}

GSurface *GreyScaleDC(GSurface *pDC)
{
	GSurface *Status = 0;

	if (pDC)
	{
		switch (pDC->GetBits())
		{
			case 8:
			{
				// 8 -> 8 greyscale convert
				Status = pDC;

				GPalette *Pal = pDC->Palette();
				if (Pal)
				{
					// convert palette
					uchar Map[256];
					ZeroObj(Map);
					for (int i=0; i<Pal->GetSize(); i++)
					{
						// Weights, r:.299, g:.587, b:.114
						GdcRGB *n = (*Pal)[i];
						if (n)
						{
							double r = ((double) n->R * 0.299) / 255;
							double g = ((double) n->G * 0.587) / 255;
							double b = ((double) n->B * 0.114) / 255;
							Map[i] = (r + g + b) * 255.0;

							n->R = n->G = n->B = i;
						}
					}

					Pal->Update();

					// convert pixels..
					for (int y=0; y<pDC->Y(); y++)
					{
						uchar *Bits = (*pDC)[y];
						for (int x=0; x<pDC->X(); x++)
						{
							Bits[x] = Map[Bits[x]];
						}
					}
				}

				break;
			}
			case 16:
			case 24:
			case 32:
			{
				Status = new GMemDC;
				if (Status AND
					Status->Create(pDC->X(), pDC->Y(), 8))
				{
					uchar RMap[256];
					uchar GMap[256];
					uchar BMap[256];
					for (int i=0; i<256; i++)
					{
						RMap[i] = i * 0.299;
						GMap[i] = i * 0.587;
						BMap[i] = i * 0.114;
					}
					
					for (int y=0; y<pDC->Y(); y++)
					{
						switch (pDC->GetBits())
						{
							case 16:
							{
								ushort *Src = (ushort*) (*pDC)[y];
								uchar *Dest = (*Status)[y];

								for (int x=0; x<pDC->X(); x++)
								{
									uchar r = R16(Src[x]) << 3;
									uchar g = G16(Src[x]) << 2;
									uchar b = B16(Src[x]) << 3;
									Dest[x] = RMap[r] + GMap[g] + BMap[b];
								}
								break;
							}
							case 24:
							{
								uchar *Src = (*pDC)[y];
								uchar *Dest = (*Status)[y];

								for (int x=0; x<pDC->X(); x++)
								{
									uchar *x3 = Src + ((x<<1) + x);
									Dest[x] = RMap[ x3[2] ] + GMap[ x3[1] ] + BMap[ x3[0] ];
								}
								break;
							}
							case 32:
							{
								ulong *Src = (ulong*) (*pDC)[y];
								uchar *Dest = (*Status)[y];

								for (int x=0; x<pDC->X(); x++)
								{
									uchar r = R32(Src[x]);
									uchar g = G32(Src[x]);
									uchar b = B32(Src[x]);
									Dest[x] = RMap[r] + GMap[g] + BMap[b];
								}
								break;
							}
						}
					}

					GPalette *Pal = new GPalette(0, 256);
					if (Pal)
					{
						// convert palette
						for (int i=0; i<Pal->GetSize(); i++)
						{
							GdcRGB *n = (*Pal)[i];
							if (n)
							{
								n->R = n->G = n->B = i;
							}
						}

						Status->Palette(Pal);
					}
				}
				break;
			}
		}
	}

	return Status;
}

bool InvertDC(GSurface *pDC)
{
	bool Status = true;

	if (pDC)
	{
		switch (pDC->GetBits())
		{
			case 8:
			{
				if (IsGreyScale(pDC))
				{
					for (int y=0; y<pDC->Y(); y++)
					{
						uchar *d = (*pDC)[y];
						if (d)
						{
							for (int x=0; x<pDC->X(); x++)
							{
								d[x] = 255 - d[x];
							}
						}
					}
					Status = true;
				}
				else
				{
					GPalette *Pal = pDC->Palette();
					if (Pal)
					{
						GdcRGB *p = (*Pal)[0];
						if (p)
						{
							for (int i=0; i<Pal->GetSize(); i++)
							{
								p[i].R = 255 - p[i].R;
								p[i].G = 255 - p[i].G;
								p[i].B = 255 - p[i].B;
							}
							Status = true;
						}
					}
				}
				break;
			}
			case 16:
			{
				for (int y=0; y<pDC->Y(); y++)
				{
					ushort *d = (ushort*) (*pDC)[y];
					if (d)
					{
						for (int x=0; x<pDC->X(); x++)
						{
							d[x] =	((31 - R16(d[x])) << 11) |
								((63 - G16(d[x])) << 5) |
								((31 - B16(d[x])));
						}
					}
				}
				break;
			}
			case 24:
			{
				for (int y=0; y<pDC->Y(); y++)
				{
					uchar *d = (*pDC)[y];
					if (d)
					{
						for (int x=0; x<pDC->X(); x++)
						{
							*d++ = 255 - *d;
							*d++ = 255 - *d;
							*d++ = 255 - *d;
						}
					}
				}
				break;
			}
			case 32:
			{
				for (int y=0; y<pDC->Y(); y++)
				{
					uchar *d = (*pDC)[y];
					if (d)
					{
						for (int x=0; x<pDC->X(); x++)
						{
							*d++ = 255 - *d;
							*d++ = 255 - *d;
							*d++ = 255 - *d;
							*d++; // leave alpha channel alone
						}
					}
				}
				break;
			}
		}
	}

	return Status;
}

bool FlipDC(GSurface *pDC, int Dir)
{
	bool Status = false;

	if (pDC)
	{
		switch (Dir)
		{
			case FLIP_X:
			{
				for (int y=0; y<pDC->Y(); y++)
				{
					for (int x=0, nx=pDC->X()-1; x<nx; x++, nx--)
					{
						COLOUR c = pDC->Get(x, y);
						pDC->Colour(pDC->Get(nx, y));
						pDC->Set(x, y);
						pDC->Colour(c);
						pDC->Set(nx, y);
					}
				}

				Status = true;
				break;
			}

			case FLIP_Y:
			{
				for (int y=0, ny=pDC->Y()-1; y<ny; y++, ny--)
				{
					for (int x=0; x<pDC->X(); x++)
					{
						COLOUR c = pDC->Get(x, y);
						pDC->Colour(pDC->Get(x, ny));
						pDC->Set(x, y);
						pDC->Colour(c);
						pDC->Set(x, ny);
					}
				}

				Status = true;
				break;
			}
		}
	}

	return Status;
}

bool RotateDC(GSurface *pDC, double Angle)
{
	GSurface *pOld = new GMemDC;
	bool Status = false;

	if (pDC AND pOld AND pOld->Create(pDC->X(), pDC->Y(), pDC->GetBits()))
	{
		// copy the palette
		// GPalette *Pal = pDC->Palette() ? new GPalette(pDC->Palette()) : 0;

		// copy the bits
		pOld->Blt(0, 0, pDC);

		// do the rotation
		if (Angle == 180)
		{
			// int Line = (pDC->X() * pDC->GetBits() + 7) / 8;

			for (int y=0; y<pOld->Y(); y++)
			{
				for (int x=0; x<pOld->X(); x++)
				{
					pDC->Colour(pOld->Get(x, y));
					pDC->Set(pOld->X()-x-1, pOld->Y()-y-1);
				}
			}

			Status = true;
		}
		else if (Angle == 90 OR Angle == 270)
		{
			if (pDC->Create(pOld->Y(), pOld->X(), pOld->GetBits()))
			{
				if (Angle == 90)
				{
					for (int y=0; y<pOld->Y(); y++)
					{
						for (int x=0; x<pOld->X(); x++)
						{
							pDC->Colour(pOld->Get(x, y));
							pDC->Set(pDC->X()-y-1, x);
						}
					}
				}
				else
				{
					for (int y=0; y<pOld->Y(); y++)
					{
						for (int x=0; x<pOld->X(); x++)
						{
							pDC->Colour(pOld->Get(x, y));
							pDC->Set(y, pDC->Y()-x-1);
						}
					}
				}

				Status = true;
			}
		}
		else
		{
			// free angle
		}
	}

	DeleteObj(pOld);

	return Status;
}

bool FlipXDC(GSurface *pDC)
{
	bool Status = false;
	if (pDC)
	{
		for (int y=0; y<pDC->Y(); y++)
		{
			int Sx = 0;
			int Ex = pDC->X()-1;
			for (; Sx<Ex; Sx++, Ex--)
			{
				COLOUR c = pDC->Get(Sx, y);
				pDC->Colour(pDC->Get(Ex, y));
				pDC->Set(Sx, y);
				pDC->Colour(c);
				pDC->Set(Ex, y);
			}
		}
		Status = true;
	}
	return Status;
}

bool FlipYDC(GSurface *pDC)
{
	bool Status = false;
	if (pDC)
	{
		for (int x=0; x<pDC->X(); x++)
		{
			int Sy = 0;
			int Ey = pDC->Y()-1;
			for (; Sy<Ey; Sy++, Ey--)
			{
				COLOUR c = pDC->Get(x, Sy);
				pDC->Colour(pDC->Get(x, Ey));
				pDC->Set(x, Sy);
				pDC->Colour(c);
				pDC->Set(x, Ey);
			}
		}
		Status = true;
	}
	return Status;
}

// Remaps DC to new palette
bool RemapDC(GSurface *pDC, GPalette *DestPal)
{
	bool Status = false;
	if (pDC AND pDC->GetBits() <= 8 AND DestPal)
	{
		GPalette *SrcPal = pDC->Palette();
		int Colours = (SrcPal) ? SrcPal->GetSize() : 1 << pDC->GetBits();
		uchar *Remap = new uchar[Colours];
		if (Remap)
		{
			if (SrcPal)
			{
				for (int i=0; i<Colours; i++)
				{
					GdcRGB *p = (*SrcPal)[i];
					if (p)
					{
						Remap[i] = DestPal->MatchRgb(Rgb24(p->R, p->G, p->B));
					}
					else
					{
						Remap[i] = 0;
					}
				}
			}
			else
			{
				for (int i=0; i<Colours; i++)
				{
					Remap[i] = DestPal->MatchRgb(Rgb24(i, i, i));
				}
			}

			for (int y=0; y<pDC->Y(); y++)
			{
				for (int x=0; x<pDC->X(); x++)
				{
					pDC->Colour(Remap[pDC->Get(x, y)]);
					pDC->Set(x, y);
				}
			}

			Status = true;
			DeleteArray(Remap);
		}
	}

	return Status;
}

/* floating point resample code...

	double Area = 0.0;
	double R = 0.0;
	double G = 0.0;
	double B = 0.0;

	// Calc the right and bottom bounds of the original scale
	double x2 = ((double) Dx+1) * ScaleX;
	double y2 = ((double) Dy+1) * ScaleY;

	// Loop through the source pixel(s) and sum all the area
	// and RGB data
	for (double x = (double) Dx * ScaleX; x<x2; x=((int)x)+1)
	{
		for (double y = (double) Dy * ScaleY; y<y2; y=((int)y)+1)
		{
			// Get the source pixel colour
			COLOUR c = CBit(24, pSrc->Get((int)x, (int)y), pSrc->GetBits(), pPal);
			
			// Calculate bounds of source pixel
			double Fx, Fy, j;
			Fx = modf(x, &j);
			double Width = (x2 < ceil(x)) ? x2 - j : 1.0;
			Fy = modf(y, &j);
			double Height = (y2 < ceil(y)) ? y2 - j : 1.0;

			// Calculate area
			double a = (Width-Fx) * (Height-Fy);
			
			// Add the colour and area to the running total
			R += a * R24(c);
			G += a * G24(c);
			B += a * B24(c);
			Area += a;
		}
	}

	R /= Area;
	G /= Area;
	B /= Area;
	COLOUR c = Rgb24( ((int)R), ((int)G), ((int)B) );
	pDest->Colour(CBit(pDest->GetBits(), c, 24));
	pDest->Set(Dx, Dy);
*/

bool ResampleDC(GSurface *pDest, GSurface *pSrc, Progress *Prog)
{
	if (pDest AND pSrc)
	{
		double ScaleX = (double) pSrc->X() / pDest->X();
		double ScaleY = (double) pSrc->Y() / pDest->Y();
		int Sx = ScaleX * 256;
		int Sy = ScaleY * 256;
		int SrcBits = pSrc->GetBits();
		GPalette *pPal = pSrc->Palette();
		int OutBits;
		
		switch (pSrc->GetBits())
		{
			case 24:
				OutBits = pDest->GetBits() == 24 ? 0 : 24;
				break;
			case 32:
				OutBits = pDest->GetBits() == 32 ? 0 : 32;
				break;
			case 8:
				if (pPal)
				{
					OutBits = 24;
				}
				else
				{
					OutBits = 0;
				}
				break;
			default:
				OutBits = 24;
				break;
		}

		if (Prog)
		{
			Prog->SetDescription("Resampling image...");
			Prog->SetLimits(0, pDest->Y()-1);
		}

		for (int Dy = 0; Dy<pDest->Y(); Dy++)
		{
			for (int Dx = 0; Dx<pDest->X(); Dx++)
			{
				int64 Area = 0;
				int64 R = 0;
				int64 G = 0;
				int64 B = 0;
				int64 A = 0;

				int NextX = (Dx+1)*Sx;
				int NextY = (Dy+1)*Sy;
				int Nx, Ny, a;
				COLOUR c;

				if (pSrc->GetBits() == 32)
				{
					for (int y=Dy*Sy; y<NextY; y=Ny)
					{
						Ny = y + (256 - (y%256));

						for (int x=Dx*Sx; x<NextX; x=Nx)
						{
							Nx = x + (256 - (x%256));

							c = pSrc->Get(x >> 8, y >> 8);
							a = (Nx - x) * (Ny - y);

							// Add the colour and area to the running total
							R += a * R32(c);
							G += a * G32(c);
							B += a * B32(c);
							A += a * A32(c);
							Area += a;
						}
					}
				}
				else if (pSrc->GetBits() == 8)
				{
					for (int y=Dy*Sy; y<NextY; y=Ny)
					{
						Ny = y + (256 - (y%256));

						for (int x=Dx*Sx; x<NextX; x=Nx)
						{
							Nx = x + (256 - (x%256));

							c = pSrc->Get(x >> 8, y >> 8);
							a = (Nx - x) * (Ny - y);

							// Add the grey and area to the running total
							G += a * c;
							Area += a;
						}
					}
				}
				else
				{
					for (int y=Dy*Sy; y<NextY; y=Ny)
					{
						Ny = y + (256 - (y%256));

						for (int x=Dx*Sx; x<NextX; x=Nx)
						{
							Nx = x + (256 - (x%256));

							c = CBit(24, pSrc->Get(x>>8, y>>8), SrcBits);
							a = (Nx - x) * (Ny - y);

							// Add the colour and area to the running total
							R += a * R24(c);
							G += a * G24(c);
							B += a * B24(c);
							Area += a;
						}
					}
				}

				if (Area)
				{
					G /= Area;
					if (pSrc->GetBits() == 8)
					{
						c = G;
					}
					else
					{
						R /= Area;
						B /= Area;
						if (pSrc->GetBits() == 32)
						{
							A /= Area;
							c = Rgba32( ((int)R), ((int)G), ((int)B), ((int)A) );
						}
						else
						{
							c = Rgb24( ((int)R), ((int)G), ((int)B) );
						}
					}
				}
				else
				{
					c = 0;
				}

				pDest->Colour(c, OutBits);
				pDest->Set(Dx, Dy);
			}

			if (Prog AND Dy % 64 == 0)
			{
				Prog->Value(Dy);
			}
		}

		return true;
	}

	return false;
}

